--[[
    Coconuts by LordFarlander
    Version 2.6.134
    $Revision: 134 $

    Used to be called Mount.
]]--

--[[
Copyright (c) 2008, LordFarlander
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--

local L = LibStub( "AceLocale-3.0" ):GetLocale( "Coconuts" );

_G["BINDING_NAME_CLICK Coconuts_Button:LeftButton"] = L["Summon Mount"];
BINDING_HEADER_COCONUTSHEADER = L["Coconuts"];

if( not LibStub( "LibLordFarlander-SpecialEvents-Mount-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-SpecialEvents-Mount-4.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-SpecialEvents-LearnSpell-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-SpecialEvents-LearnSpell-4.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-SpecialEvents-Skill-4.0" ) );
    return;
end--if

local LibLordFarlander = LibStub( "LibLordFarlander-2.0", true );
if( not LibLordFarlander ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-2.0" ) );
    return;
end--if

local ItemList = LibStub( "LibLordFarlander-ItemList-2.0", true );
if( not ItemList ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-ItemList-2.0" ) );
    return;
end--if

if( not LibStub( "LibLordFarlander-UI-2.0", true ) ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-UI-2.0" ) );
    return;
end--if

local PetAndMountDatabase = LibStub( "LibLordFarlander-PetAndMountDatabase-1.1", true );
if( not PetAndMountDatabase ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-PetAndMountDatabase-1.1" ) );
    return;
end--if

local TZ = LibStub( "LibLordFarlander-TableRecycler-1.0", true );
if( not TZ ) then
    error( L["%s requires %s"]:format( L["Coconuts"], "LibLordFarlander-TableRecycler-1.0" ) );
    return;
end--if

Coconuts = LibStub( "AceAddon-3.0" ):NewAddon( "Coconuts", "AceEvent-3.0", "AceTimer-3.0", "AceConsole-3.0" );
Coconuts.version = { major = 2, minor = 6, revision = tonumber( "134" ) };

Coconuts.DEBUG  = false;            --true = Debugging Mode
Coconuts.dprint = LibLordFarlander.dprint;
Coconuts.button = nil;
Coconuts.ui = { Favorites = {}, };

Coconuts.dirtyBags = { [0] = true, [1] = true, [2] = true, [3] = true, [4] = true, [-254] = true, [-255] = true, };

--DO NOT CHANGE THESE VALUES
Coconuts.Ground = { Best = "", NextBest = "", Passenger = { Best = "", NextBest = "", Sets = { [""] = {}, }, }, Sets = { [""] = {}, }, };
Coconuts.Flying = { Best = "", NextBest = "", Passenger = { Best = "", NextBest = "", Sets = { [""] = {}, }, }, Sets = { [""] = {}, }, };

Coconuts.CurrentGround     = nil; --Ground mount on the Button
Coconuts.CurrentFlying     = nil; --Flying mount on the Button
Coconuts.CurrentMount      = nil; --Current mount
Coconuts.CurrentMountItem  = nil; --Current mount modifier item
Coconuts.NextCurrentGround = nil; --Ground mount on the Button
Coconuts.NextCurrentFlying = nil; --Flying mount on the Button
Coconuts.CurrentGroundForm = nil;
Coconuts.CurrentFlyingForm = nil;
Coconuts.CurrentAquaticForm = nil;
Coconuts.CurrentGroundPassenger     = nil;
Coconuts.CurrentFlyingPassenger     = nil;
Coconuts.NextCurrentGroundPassenger = nil;
Coconuts.NextCurrentFlyingPassenger = nil;
Coconuts.ReadyToInitialize = false;
Coconuts.Mounted           = false;
Coconuts.Swimming          = false;
Coconuts.InVehicle         = false;
Coconuts.Dirty             = false;
Coconuts.Mode              = "";
Coconuts.RidingSkill       = nil;
Coconuts.Loaded            = false;

Coconuts.MountItemNames = {
    [21212] = L["Fresh Holly"],
    [37750] = L["Fresh Brewfest Hops"],
    [39477] = L["Fresh Dwarven Hops"],
    [39476] = L["Fresh Goblin Hops"],
    [21213] = L["Preserved Holly"],
    [37816] = L["Preserved Brewfest Hops"],
};

function Coconuts:OnInitialize()
    local AceConfigReg = LibStub( "AceConfigRegistry-3.0" );
    local AceConfigDialog = LibStub( "AceConfigDialog-3.0" );
    local defaults = {
            profile = {
                FavorFlying = true,
                FavorPassenger = true,
                ShowModifierItems = true,
                SuggestOnlySoulbound = true,
                ModiferDismountFlying = true,
                SuggestForms = true,
                FlightFormDefault = true,
                DefaultToFormInCombat = true,
                FavorMorePassengers = false,
                FlyingModifier = "Alt",
                NextBestModifier = "Shift",
                PassengerModifier = "Control",
                MountItemModifier = "Shift",
                -- List of favorite mounts, these will be picked more often
                -- stored as [itemID]=percentage
                Favorites = {},
                -- List of mounts per zone, if the mounts are in the player's bags or spells they will be picked
                -- when in the indicated zone
                -- stored as ["Localized zone name"]={ "type"=itemID, ... }
                ZoneFavorites = {},
                Button = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonSettingsDefaults(),
            },
        };
    local buttonOptions = LibStub( "LibLordFarlander-UI-2.0" ).GetButtonOptionGroup( self, {
            name = L["Button"],
            desc = L["Options relating to the button."],
            yes = L["yes"],
            no = L["no"],
            show = L["showing"],
            hide = L["hiding"],
            locked = L["locked"],
            unlocked = L["unlocked"],
            toggleButton = L["toggleButton"],
            toggleButton_name = L["Show button"],
            toggleButton_desc = L["Show/hide the button."],
            autoHideButton = L["autoHideButton"],
            autoHideButton_name = L["Automatically hide button"],
            autoHideButton_desc = L["Hide the button when you can't mount?"],
            lockButton = L["lockButton"],
            lockButton_name = L["Lock button position"],
            lockButton_desc = L["Lock/unlock the position of the button."],
            alpha = L["alpha"],
            alpha_name = L["Alpha of button"],
            alpha_desc = L["Sets the alpha value that the button has normally."],
            fadeOut = L["fadeOut"],
            fadeOut_name = L["Fade out"],
            fadeOut_desc = L["Fade the button out when the cursor is not over it?"],
            fadeOutInsteadHide = L["fadeOutInsteadHide"],
            fadeOutInsteadHide_name = L["Fade out instead of hiding"],
            fadeOutInsteadHide_desc = L["Fade the button out instead of hiding it?"],
            fadeOutCancelInCombat = L["fadeOutCancelInCombat"],
            fadeOutCancelInCombat_name = L["Cancel fade out if in combat"],
            fadeOutCancelInCombat_desc = L["Should the button not fade out if you're in combat?"],
            fadeOutCancelOnShift = L["fadeOutCancelOnShift"],
            fadeOutCancelOnShift_name = L["Cancel fade out if holding Shift"],
            fadeOutCancelOnShift_desc = L["Should the button not fade out if you're holding Shift?"],
            fadeOutCancelOnCtrl = L["fadeOutCancelOnCtrl"],
            fadeOutCancelOnCtrl_name = L["Cancel fade out if holding Ctrl"],
            fadeOutCancelOnCtrl_desc = L["Should the button not fade out if you're holding Ctrl?"],
            fadeOutCancelOnAlt = L["fadeOutCancelOnAlt"],
            fadeOutCancelOnAlt_name = L["Cancel fade out if holding Alt"],
            fadeOutCancelOnAlt_desc = L["Should the button not fade out if you're holding Alt?"],
            fadeOutTime = L["fadeOutTime"],
            fadeOutTime_name = L["Fade out time"],
            fadeOutTime_desc = L["Sets how long the button takes to fade out."],
            fadeOutDelay = L["fadeOutDelay"],
            fadeOutDelay_name = L["Fade out delay"],
            fadeOutDelay_desc = L["Sets the delay before the button starts to fade."],
            fadeOutAlpha = L["fadeOutAlpha"];
            fadeOutAlpha_name = L["Alpha after fadeout"],
            fadeOutAlpha_desc = L["Sets the alpha value that the button has after faded out."],
            buttonScale = L["buttonScale"],
            buttonScale_name = L["Scale of button"],
            buttonScale_desc = L["Sets the scaling value for the button to make it bigger or smaller."],
        }, 1 );
    local modifierItemsOptions = {
        type = "group",
        name = L["Modifier Items"],
        desc = L["Allows you to set the order that modifier items are picked."],
        handler = self,
        args = {
            favorites = {
                type = "select",
                order = 1,
                name = L["Modifier Items"],
                desc = L["Allows you to set the order that modifier items are picked."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             local favorites = {};

                             for _, item in pairs( info.handler.db.profile.enchantOrder ) do
                                 local sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( item );

                                 if( sTexture ) then
                                     tinsert( favorites, ("|T%s:0|t %s"):format( sTexture, sName ) );
                                 else
                                     tinsert( favorites, info.handler.MountItemNames[item] or item );
                                 end--if
                             end--if
                             if( not info.handler.ui.currentModifierItem ) then
                                 info.handler.ui.currentModifierItem = 1;
                             end--if
                             return favorites;                             
                         end,
                get = function( info )
                          return info.handler.ui.currentModifierItem;
                      end,
                set = function( info, val )
                          info.handler.ui.currentModifierItem = val;
                      end,
            },
            up = {
                type = "execute",
                order = 2,
                name = L["Move up"],
                desc = L["Moves the selected modifier up in the list."],
                cmdHidden = true,
                func = function( info )
                           if( info.handler.ui.currentModifierItem > 1 ) then
                               local olditem = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem - 1];
                               
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem - 1] = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem];
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem] = olditem;
                               info.handler.ui.currentModifierItem = info.handler.ui.currentModifierItem - 1;
                               info.handler:FindNextMountEnchant();
                           end--if
                       end,
            },
            down = {
                type = "execute",
                order = 2,
                name = L["Move down"],
                desc = L["Moves the selected modifier down in the list."],
                cmdHidden = true,
                func = function( info )
                           if( info.handler.ui.currentModifierItem < #info.handler.db.profile.enchantOrder ) then
                               local olditem = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem + 1];
                               
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem + 1] = info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem];
                               info.handler.db.profile.enchantOrder[info.handler.ui.currentModifierItem] = olditem;
                               info.handler.ui.currentModifierItem = info.handler.ui.currentModifierItem + 1;
                               info.handler:FindNextMountEnchant();                               
                           end--if
                       end,
            },
        },
    };
    local favoritesOptions = {
        type = "group",
        name = L["Favorites"],
        desc = L["Mounts that should be picked more often if you have them."],
        handler = self,
        args = {
            favorites = {
                type = "select",
                order = 1,
                name = L["Favorites"],
                desc = L["Mounts that should be picked more often if you have them."],
                width = "full",
                cmdHidden = true,
                values = function( info )
                             TZ.ClearTable( info.handler.ui.Favorites );

                             local Favorites = info.handler.ui.Favorites;
                             local firstItem = nil;
                             local used = TZ:GetTable();

                             for _, mount in pairs( info.handler:GetAllMountsList() ) do
                                 local id = mount:GetID();

                                 used[id] = true;
                                 table.insert( Favorites, id );
                                 if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                     info.handler.db.profile.Favorites[id] = 2.0;
                                 end--if
                                 if( ( type( info.handler.db.profile.Favorites[id] ) ~= "nil" ) and ( info.handler.db.profile.Favorites[id] == 1.0 ) ) then
                                     info.handler.db.profile.Favorites[id] = nil;
                                 end--if
                             end--for
                             for id, howOften in pairs( info.handler.db.profile.Favorites ) do
                                 if( not used[id] and ( ( id < -1 ) and GetSpellInfo( id * -1 ) or GetItemInfo( id ) ) ) then
                                     table.insert( Favorites, id );
                                     used[id] = true;
                                 end--if
                             end--if
                             TZ:ReturnTable( used );
                             table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                             Favorites = {};
                             for index, mount in pairs( info.handler.ui.Favorites ) do
                                 local sName, sTexture;
                                 local sHowOften = L["default"];
                                 local howOften = info.handler.db.profile.Favorites[mount] or 1.0;

                                 if( mount < 0 ) then
                                     sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                                 else
                                     sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                                 end--if
                                 if( howOften == 0.0 ) then
                                     sHowOften = L["never"];
                                 elseif( howOften ~= 1.0 ) then
                                     sHowOften = L["%.1f times as often"]:format( howOften );
                                 end--if                                 
                                 if( sTexture ) then
                                     Favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften );
                                 end--if
                                 if( firstItem == nil ) then
                                     firstItem = index;
                                 end--if
                             end--for
                             if( ( not Favorites[info.handler.ui.currentFavorite] ) or ( info.handler.ui.currentFavorite == nil ) ) then
                                 info.handler.ui.currentFavorite = firstItem;
                             end--if
                             return Favorites;
                         end,
                get = function( info )
                          return info.handler.ui.currentFavorite;
                      end,
                set = function( info, val )
                          info.handler.ui.currentFavorite = val;
                      end,
            },
            frequency = {
                type = "range",
                order = 2,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 11.0,
                bigStep = 0.1,
                isPercent = true,
                desc = L["Set how often the mount gets picked.  2 is twice as often and 0 is never."];
                get = function( info )
                          return info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] or 1.0;
                      end,
                set = function( info, val )
                          if( val == 1.0 ) then
                              info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          else
                              info.handler.db.profile.Favorites[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = val;
                          end--if
                      end,
            },

            [L["list"]] = {
                type = "execute",
                order = 3,
                guiHidden = true,
                name = L["List all favorite mounts"],
                desc = L["Show a list with all of your favorite mounts."],
                func = function( info )
                           info.handler:Print( L["List of your favorite mounts:"] );
                           for mount, howOften in pairs( info.handler.db.profile.Favorites ) do
                               if( howOften ~= 1.0 ) then
                                   local link = ( mount > 0 ) and LibLordFarlander.GetItemLink( mount ) or GetSpellLink( mount * -1 );

                                   if( mount > 0 ) then
                                       sTexture = select( 10, GetItemInfo( mount ) );
                                   else
                                       sTexture = select( 3, GetSpellInfo( mount * -1 ) );
                                   end--if
                                   if( howOften == 0.0 ) then
                                       howOften = L["never"];
                                   else
                                       howOften = L["%.1f times as often"]:format( howOften );
                                   end--if

                                   if( sTexture ) then
                                       info.handler:Print( ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( mount ), howOften ) );
                                   else
                                       info.handler:Print( ("%s"):format( link or tostring( mount ) ) );
                                   end--if
                               end--if
                           end--for
                       end,
            },
        },
    };
    local typeTemplate = {
        header = {
            type = "header",
            order = 1,
            width = "full",
            cmdHidden = true,
            name = function( info )
                       local setname = LibLordFarlander.UnnamifyString( info[#info - 1] );
                       local speed = self.GetSetSpeed( setname );
                       local type;

                       if( setname:find( "%.Passenger" ) ) then
                           if( setname:find( "%.Flying" ) ) then
                               type = L["flying passenger"];
                           else
                               type = L["ground passenger"];
                           end--if
                       else
                           if( setname:find( "%.Flying" ) ) then
                               type = L["flying"];
                           else
                               type = L["ground"];
                           end--if
                       end--if
                       return ("%s: %s"):format( LibLordFarlander.UnnamifyString( info[#info - 2] ), L["%1$i%% speed %2$s"]:format( speed, type ) );
                   end,
        },
        mount = {
            type = "select",
            order = 2,
            name = L["Mount"],
            desc = L["Mounts that should be picked more often if you have them in the selected zone and type."],
            width = "full",
            cmdHidden = true,
            values = function( info )
                         local currentType = LibLordFarlander.UnnamifyString( info[#info - 1] ) or "";
                         local currentZone = LibLordFarlander.UnnamifyString( info[#info - 2] );

                         TZ.ClearTable( info.handler.ui.Favorites );
                         if( not info.handler.db.profile.ZoneFavorites[currentZone] ) then
                             info.handler.db.profile.ZoneFavorites[currentZone] = {};
                         end--if

                         local Favorites = info.handler.ui.Favorites;
                         local used = TZ:GetTable();
                         local ZoneFavorite = info.handler.db.profile.ZoneFavorites[currentZone][currentType];

                         for _, mount in pairs( info.handler:GetPetsList( currentType ) ) do
                             local id = mount:GetID();
                             table.insert( Favorites, id );
                             used[id] = true;
                             if( type( info.handler.db.profile.Favorites[id] ) == "boolean" ) then
                                 info.handler.db.profile.Favorites[id] = 2.0;
                             end--if
                             if( ( type( info.handler.db.profile.Favorites[id] ) ~= "nil" ) and ( info.handler.db.profile.Favorites[id] == 1.0 ) ) then
                                 info.handler.db.profile.Favorites[id] = nil;
                             end--if                             
                         end--for
                         if( ZoneFavorite ) then
                             if( type( ZoneFavorite ) ~= "table" ) then
                                 info.handler.db.profile.ZoneFavorites[currentZone][currentType] = { [ZoneFavorite] = { frequency = 1.0, exclusive = true } };
                                 ZoneFavorite = info.handler.db.profile.ZoneFavorites[currentZone][currentType];
                             end--if
                             for mount, _ in pairs( ZoneFavorite ) do
                                 if( not used[mount] and ( ( mount < -1 ) and GetSpellInfo( mount * -1 ) or GetItemInfo( mount ) ) ) then
                                     table.insert( Favorites, mount );
                                     used[mount] = true;
                                 end--if
                             end--for
                         else
                             info.handler.db.profile.ZoneFavorites[currentZone][currentType] = {};
                             ZoneFavorite = info.handler.db.profile.ZoneFavorites[currentZone][currentType];
                         end--if
                         TZ:ReturnTable( used );
                         table.sort( Favorites, function( a, b ) return ( ( a < 0 ) and GetSpellInfo( a * -1 ) or GetItemInfo( a ) ) < ( ( b < 0 ) and GetSpellInfo( b * -1 ) or GetItemInfo( b ) ) end );
                         Favorites = {};
                         for index, mount in pairs( info.handler.ui.Favorites ) do
                             local sName, sTexture;
                             local howOften = 1.0;
                             local sHowOften = L["default"]
                             local exclusive = false;

                             if( ZoneFavorite[mount] ) then
                                 howOften = ZoneFavorite[mount].frequency or 1.0;
                                 exclusive = ZoneFavorite[mount].exclusive;
                             end--if
                             if( mount < 0 ) then
                                 sName, _, sTexture, _, _, _, _, _, _ = GetSpellInfo( mount * -1 );
                             else
                                 sName, _, _, _, _, _, _, _, _, sTexture = GetItemInfo( mount );
                             end--if
                             if( howOften == 0.0 ) then
                                 sHowOften = L["never"];
                             elseif( howOften ~= 1.0 ) then
                                 sHowOften = L["%.1f times as often"]:format( howOften );
                             end--if
                             if( exclusive ) then
                                 sHowOften = string.format( "%s, %s", sHowOften, L["exclusive"] );
                             end--if
                             if( sTexture ) then
                                 Favorites[index] = ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, sName, sHowOften );
                             end--if
                         end--for
                         if( #Favorites == 0 ) then
                             info.handler.ui.currentFavorite = nil;
                             Favorites[0] = L["none"];
                         elseif( ( info.handler.ui.currentFavorite == nil ) or ( info.handler.ui.currentFavorite > #Favorites ) ) then
                             info.handler.ui.currentFavorite = 1;
                         end--if
                         return Favorites;
                     end,
            get = function( info )
                      return info.handler.ui.currentFavorite;
                  end,
            set = function( info, val )
                      info.handler.ui.currentFavorite = val;
                  end,
            },
            frequency = {
                type = "range",
                order = 3,
                name = L["Frequency"],
                cmdHidden = true,
                min = 0.0,
                max = 11.0,
                bigStep = 0.1,
                isPercent = true,
                desc = L["Set how often the mount gets picked.  2 is twice as often and 0 is never."];
                get = function( info )
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 2] )];
                          
                          if( ZoneFavorite ) then
                              local ZoneType = ZoneFavorite[LibLordFarlander.UnnamifyString( info[#info - 1] )];

                              if( ZoneType and ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] ) then
                                  return ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]].frequency or 1.0;
                              end--if
                          end--if
                          return 1.0;
                      end,
                set = function( info, val )
                          local zone = LibLordFarlander.UnnamifyString( info[#info - 2] );
                          local type = LibLordFarlander.UnnamifyString( info[#info - 1] );
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                          local ZoneType;
                          local CurrentFavorite;

                          if( val == 1.0 ) then
                              val = nil;
                          end--if
                          if( not ZoneFavorite ) then
                              info.handler.db.profile.ZoneFavorites[zone] = {};
                              ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                              ZoneFavorite[type] = {};
                          end--if
                          ZoneType = ZoneFavorite[type];
                          if( not ZoneType ) then
                              ZoneFavorite[type] = {};
                              ZoneType = ZoneFavorite[type];
                          end--if
                          CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]];
                          if( not CurrentFavorite ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = { frequency = val };
                              CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]]
                          else
                              CurrentFavorite.frequency = val;
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          end--if
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            active = {
                type = "toggle",
                order = 4,
                name = L["Override favorites"],
                desc = L["Check this to override the favorites settings for this mount."];
                get = function( info )
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 2] )];
                          
                          if( ZoneFavorite ) then
                              local ZoneType = ZoneFavorite[LibLordFarlander.UnnamifyString( info[#info - 1] )];
                              
                              if( ZoneType and ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] ) then
                                  return ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]].active or false;
                              end--if
                          end--if
                          return false;
                      end,
                set = function( info, val )
                          local zone = LibLordFarlander.UnnamifyString( info[#info - 2] );
                          local type = LibLordFarlander.UnnamifyString( info[#info - 1] );
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                          local ZoneType;
                          local CurrentFavorite;

                          if( not ZoneFavorite ) then
                              info.handler.db.profile.ZoneFavorites[zone] = {};
                              ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                              ZoneFavorite[type] = {};
                          end--if
                          ZoneType = ZoneFavorite[type];
                          if( not ZoneType ) then
                              ZoneFavorite[type] = {};
                              ZoneType = ZoneFavorite[type];
                          end--if
                          CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]];
                          if( not CurrentFavorite ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = { active = val };
                              CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]]
                          else
                              CurrentFavorite.active = val;
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          end--if
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            exclusive = {
                type = "toggle",
                order = 5,
                name = L["Exclusive"],
                desc = L["Mounts marked Exclusive will be the only ones picked from."];
                get = function( info )
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[LibLordFarlander.UnnamifyString( info[#info - 2] )];

                          if( ZoneFavorite ) then
                              local ZoneType = ZoneFavorite[LibLordFarlander.UnnamifyString( info[#info - 1] )];

                              if( ZoneType and ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] ) then
                                  return ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]].exclusive or false;
                              end--if
                          end--if
                          return false;
                      end,
                set = function( info, val )
                          local zone = LibLordFarlander.UnnamifyString( info[#info - 2] );
                          local type = LibLordFarlander.UnnamifyString( info[#info - 1] );
                          local ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                          local ZoneType;
                          local CurrentFavorite;

                          if( not ZoneFavorite ) then
                              info.handler.db.profile.ZoneFavorites[zone] = {};
                              ZoneFavorite = info.handler.db.profile.ZoneFavorites[zone];
                              ZoneFavorite[type] = {};
                          end--if
                          ZoneType = ZoneFavorite[type];
                          if( not ZoneType ) then
                              ZoneFavorite[type] = {};
                              ZoneType = ZoneFavorite[type];
                          end--if
                          CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]];
                          if( not CurrentFavorite ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = { exclusive = val };
                              CurrentFavorite = ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]]
                          else
                              CurrentFavorite.exclusive = val;
                          end--if
                          if( not ( CurrentFavorite.frequency or CurrentFavorite.active or CurrentFavorite.exclusive ) ) then
                              ZoneType[info.handler.ui.Favorites[info.handler.ui.currentFavorite]] = nil;
                          end--if
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
    };
    local zoneTemplate = {
    };
    local continentTemplate = {
    };
    local zoneOptions =  {
        type = "group",
        name = L["Zone Favorites"],
        desc = L["Set a mount to use when in a certain zone."],
        handler = self,
        childGroups = "tree",
        args = {
            [L["list"]] = {
                type = "execute",
                order = -1,
                guiHidden = true,
                name = L["List all zone favorite mounts"],
                desc = L["Show a list with all of your zone favorite mounts."],
                func = function( info )
                           info.handler:Print( L["List of your zone favorite mounts:"] );
                           for zone, types in pairs( info.handler.db.profile.ZoneFavorites ) do
                               for type, mounts in pairs( types ) do
                                   for mount, settings in pairs( mounts ) do
                                       local howOften = settings.frequency or 1.0;
                                       local sHowOften =  "default";
                                       local exclusive = settings.exclusive;
    
                                       if( ( howOften ~= 1.0 ) or exclusive ) then
                                           local link = ( mount > 0 ) and LibLordFarlander.GetItemLink( mount ) or GetSpellLink( mount * -1 );
        
                                           if( mount > 0 ) then
                                               sTexture = select( 10, GetItemInfo( mount ) );
                                           else
                                               sTexture = select( 3, GetSpellInfo( mount * -1 ) );
                                           end--if
                                           if( howOften == 0.0 ) then
                                               sHowOften = L["never"];
                                           elseif( howOften ~= 1.0 ) then
                                               sHowOften = L["%.1f times as often"]:format( howOften );
                                           end--if
                                           if( exclusive ) then
                                               sHowOften = ("%s, %s"):format( sHowOften, L["exclusive"] );
                                           end--if
                                           if( sTexture ) then
                                               info.handler:Print( ("|T%s:0|t %s |cFFFF0000[%s]|r"):format( sTexture, link or tostring( mount ), sHowOften ) );
                                           else
                                               info.handler:Print( ("%s |cFFFF0000[%s]|r"):format( link or tostring( mount ), sHowOften ) );
                                           end--if
                                       end--if
                                   end--for
                               end--for
                           end--for
                       end,
            },
        },
    };
    local modifierKeys = { ["Disabled"] = L["Disabled"], ["Alt"] = L["Alt"], ["Control"] = L["Control"], ["Shift"] = L["Shift"], };
    local options = {
        type = "group",
        handler = self,
        args = {
            [L["favorFlying"]] = {
                type = "toggle",
                order = 1,
                name = L["Favor flying mounts"],
                desc = function( info )
                            local returnString = L["Toggles if flying mounts should be favored if in an area that allows them."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.FavorFlying and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.FavorFlying;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorFlying = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["favorPassenger"]] = {
                type = "toggle",
                order = 2,
                name = L["Favor passenger mounts"],
                desc = function( info )
                            local returnString = L["Toggles if mounts that can take a passenger should be favored."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.FavorPassenger and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.FavorPassenger;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorPassenger = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["favorMorePassengers"]] = {
                type = "toggle",
                order = 3,
                name = L["Favor more passengers"],
                desc = function( info )
                            local returnString = L["Toggles if mounts that can take more passengers should be favored over mounts that can take less."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.FavorPassenger and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.FavorMorePassengers;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FavorMorePassengers = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["showModifierItems"]] = {
                type = "toggle",
                order = 4,
                name = L["Show modifier items"],
                desc = function( info )
                            local returnString = L["Toggles if items that modify mounts (such as Fresh Holly) should be selected when holding Ctrl while mounted."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ShowModifierItems and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.ShowModifierItems;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ShowModifierItems = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["suggestOnlySoulbound"]] = {
                type = "toggle",
                order = 5,
                name = L["Suggest only soulbound mounts"],
                desc = function( info )
                            local returnString = L["Toggles if only mounts that are soulbound to you are to be suggested for use."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.SuggestOnlySoulbound and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.SuggestOnlySoulbound;
                      end,
                set = function( info, val )
                          info.handler.db.profile.SuggestOnlySoulbound = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["modiferDismountFlying"]] = {
                type = "toggle",
                order = 6,
                name = L["Need modifer to dismount while flying"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if a modifier needs to be held down to dismount while flying."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.ModiferDismountFlying and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.ModiferDismountFlying;
                      end,
                set = function( info, val )
                          info.handler.db.profile.ModiferDismountFlying = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["suggestForms"]] = {
                type = "toggle",
                order = 7,
                name = L["Suggest travel forms instead of next best mount"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if a travel form spell will be selected instead of your next slowest mount, if availiable."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.SuggestForms and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.SuggestForms;
                      end,
                set = function( info, val )
                          info.handler.db.profile.SuggestForms = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["flightFormDefault"]] = {
                type = "toggle",
                order = 8,
                name = L["Default to flight form"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if flight form will be defaulted to instead of a flying mount."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.SuggestForms and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.FlightFormDefault;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FlightFormDefault = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            [L["defaultToFormInCombat"]] = {
                type = "toggle",
                order = 9,
                name = L["Suggest travel form when in combat"],
                width = "double",
                desc = function( info )
                            local returnString = L["Toggles if when entering combat your travel form spell will be selected."];

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.db.profile.DefaultToFormInCombat and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.db.profile.DefaultToFormInCombat;
                      end,
                set = function( info, val )
                          info.handler.db.profile.DefaultToFormInCombat = val;
                          info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                      end,
            },
            flyingModifier = {
                type = "select",
                order = 10,
                name = L["Modifier key for flying mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non flying mount in an area where flying is allowed."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.FlyingModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.FlyingModifier = val;
                      end,
            },
            passengerModifier = {
                type = "select",
                order = 11,
                name = L["Modifier key for passenger mounts"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a non passenger mount."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.PassengerModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.PassengerModifier = val;
                      end,
            },
            mountItemModifier = {
                type = "select",
                order = 12,
                name = L["Modifier key for modifier items"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick a modifer item while mounted."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.MountItemModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.MountItemModifier = val;
                      end,
            },
            nextBestModifier = {
                type = "select",
                order = 13,
                name = L["Modifier key for next slowest mount"],
                values = modifierKeys,
                cmdHidden = true,
                desc = L["Changes the key that needs to be held down to pick your next slowest mount."],
                width = "double",
                get = function( info )
                          return info.handler.db.profile.NextBestModifier;
                      end,
                set = function( info, val )
                          info.handler.db.profile.NextBestModifier = val;
                      end,
            },
            [L["list"]] = {
                type = "execute",
                order = 14,
                name = L["List all mounts"],
                desc = L["Show a list with all your mounts."],
                cmdHidden = true,
                func = function( info )
                           info.handler:Print( L["List of your mounts:"] );

                           info.handler:ListMountType( L["ground"], info.handler.Ground.Sets );
                           info.handler:ListMountType( L["flying"], info.handler.Flying.Sets );
                           info.handler:ListMountType( L["ground passenger"], info.handler.Ground.Passenger.Sets );
                           info.handler:ListMountType( L["flying passenger"], info.handler.Flying.Passenger.Sets );
                       end,
            },
            [L["refresh"]] = {
                type = "execute",
                order = 15,
                name = L["Refresh mount list"],
                desc = L["Force a refresh of the list of all your mounts."],
                cmdHidden = true,
                func = function( info )
                           info.handler.dirtyBags[0] = true;
                           info.handler.dirtyBags[1] = true;
                           info.handler.dirtyBags[2] = true;
                           info.handler.dirtyBags[3] = true;
                           info.handler.dirtyBags[4] = true;
                           info.handler.dirtyBags[-254] = true;
                           info.handler.dirtyBags[-255] = true;
                           info.handler:GetPlayersMounts();
                           info.handler:ChooseRandomMounts();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
        },
        name = L["Coconuts"],
        desc = L["Addon options."],
        icon = "Interface\\AddOns\\Coconuts\\Img\\Coconuts",
    };

    self.db = LibStub( "AceDB-3.0" ):New( "CoconutsDB", defaults );
    self.db.RegisterCallback( self, "OnProfileChanged" );
    self.db.RegisterCallback( self, "OnProfileReset" );

    local profileOptions = LibStub( "AceDBOptions-3.0" ):GetOptionsTable( self.db );
    local slashOptions = {
        type = "group",
        handler = self,
        name = options.name,
        desc = options.desc,
        icon = options.icon,
        args = {
            [L["options"]] = options,
            [L["favorites"]] = favoritesOptions,
            [L["zoneFavorites"]] = zoneOptions,
            [L["button"]] = buttonOptions,
            [L["profile"]] = profileOptions,
            [L["gui"]] = {
                type = "execute",
                order = -1,
                name = L["Open GUI configurator"],
                desc = L["Open GUI configurator"],
                guiHidden = true,
                func = function( info )
                           InterfaceOptionsFrame_OpenToCategory( L[info.handler.name] );
                       end,
            },
            [L["newRandom"]] = {
                type = "execute",
                order = 6,
                name = L["Force the picking of new random mounts"],
                desc = L["Force the picking of new random mounts"],
                guiHidden = true,
                func = function( info )
                           info.handler:ChooseRandomMounts();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
            [L["nextItem"]] = {
                type = "execute",
                order = 7,
                name = L["Cycle to the next mount modifier item"],
                desc = L["Cycle to the next mount modifier item"],
                guiHidden = true,
                func = function( info )
                           info.handler:FindNextMountEnchant();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
            [L["summon"]] = {
                type = "group",
                order = 8,
                name = L["Summoning mounts"],
                desc = L["Summoning mounts"],
                guiHidden = true,
                args = {
                    [L["best"]] = {
                        type = "execute",
                        order = 1,
                        name = L["Summons the current best random mount (if a learned companion)"],
                        desc = L["Summons the current best random mount (if a learned companion)"],
                        guiHidden = true,
                        func = function( info )
                                   local mount = nil;
                                   local CurrentMountIndex = nil;

                                   -- logic to find the best mount
                                   if( LibLordFarlander.IsInFlyableZone() or ( ( GetRealZoneText() == BZ["Dalaran"] ) and ( GetSubZoneText() == L["Krasus' Landing"] ) and LibLordFarlander.HasPlayerSpell( LibLordFarlander.ColdWeatherFlying ) ) ) then
                                       if( profile.FavorFlying ) then
                                           if( profile.FavorPassenger ) then
                                               if( self.CurrentFlyingPassenger ) then
                                                   mount = self.CurrentFlyingPassenger;
                                               end--if
                                           end--if
                                           if( self.CurrentFlying ) then
                                               mount = self.CurrentFlying;
                                           end--if
                                       end--if
                                   end--if
                                   if( profile.FavorPassenger ) then
                                       if( self.CurrentGroundPassenger ) then
                                           mount = self.CurrentGroundPassenger;
                                       end--if
                                   end--if
                                   if( self.CurrentGround ) then
                                       mount = self.CurrentGround;
                                   end--if
                                   if( mount ) then
                                       info.handler:QuickSummon( mount.spellID );
                                   end--if
                               end,
                    },                
                    [L["ground"]] = {
                        type = "group",
                        order = 2,
                        name = L["Ground mounts"],
                        desc = L["Ground mounts"],
                        guiHidden = true,
                        args = {
                            [L["best"]] = {
                                type = "execute",
                                order = 1,
                                name = L["Summons the current best random mount (if a learned companion)"],
                                desc = L["Summons the current best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentGround ) then
                                               info.handler:QuickSummon( info.handler.CurrentGround.spellID );
                                           end--if
                                       end,
                            },
                            [L["nextbest"]] = {
                                type = "execute",
                                order = 2,
                                name = L["Summons the current next best random mount (if a learned companion)"],
                                desc = L["Summons the current next best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.NextCurrentGround ) then
                                               info.handler:QuickSummon( info.handler.NextCurrentGround.spellID );
                                           end--if
                                       end,
                            },
                        },
                    },
                    [L["flying"]] = {
                        type = "group",
                        order = 3,
                        name = L["Flying mounts"],
                        desc = L["Flying mounts"],
                        guiHidden = true,
                        args = {
                            [L["best"]] = {
                                type = "execute",
                                order = 1,
                                name = L["Summons the current best random mount (if a learned companion)"],
                                desc = L["Summons the current best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentFlying ) then
                                               info.handler:QuickSummon( info.handler.CurrentFlying.spellID );
                                           end--if
                                       end,
                            },
                            [L["nextbest"]] = {
                                type = "execute",
                                order = 2,
                                name = L["Summons the current next best random mount (if a learned companion)"],
                                desc = L["Summons the current next best random mount (if a learned companion)"],
                                guiHidden = true,
                                func = function( info )
                                           if( info.handler.CurrentFlying ) then
                                               info.handler:QuickSummon( info.handler.CurrentFlying.spellID );
                                           end--if
                                       end,
                            },                        
                        },
                    },
                    [L["passenger"]] = {
                        type = "group",
                        order = 4,
                        name = L["Passenger mounts"],
                        desc = L["Passenger mounts"],
                        guiHidden = true,
                        args = {
                            [L["ground"]] = {
                                type = "group",
                                order = 2,
                                name = L["Ground mounts"],
                                desc = L["Ground mounts"],
                                guiHidden = true,
                                args = {
                                    [L["best"]] = {
                                        type = "execute",
                                        order = 1,
                                        name = L["Summons the current best random mount (if a learned companion)"],
                                        desc = L["Summons the current best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.CurrentPassengerGround ) then
                                                       info.handler:QuickSummon( info.handler.CurrentPassengerGround.spellID );
                                                   end--if                                        
                                               end,
                                    },
                                    [L["nextbest"]] = {
                                        type = "execute",
                                        order = 2,
                                        name = L["Summons the current next best random mount (if a learned companion)"],
                                        desc = L["Summons the current next best random mount (if a learned companion)"],
                                        guiHidden = true,
                                        func = function( info )
                                                   if( info.handler.NextCurrentPassengerGround ) then
                                                       info.handler:QuickSummon( info.handler.NextCurrentPassengerGround.spellID );
                                                   end--if
                                               end,
                                    },
                                },
                            },                        
                        },
                    },
                },
            },
            [L["list"]] = {
                type = "execute",
                order = 14,
                name = L["List all mounts"],
                desc = L["Show a list with all your mounts."],
                guiHidden = true,
                func = function( info )
                           info.handler:Print( L["List of your mounts:"] );

                           info.handler:ListMountType( L["ground"], info.handler.Ground.Sets );
                           info.handler:ListMountType( L["flying"], info.handler.Flying.Sets );
                           info.handler:ListMountType( L["ground passenger"], info.handler.Ground.Passenger.Sets );
                           info.handler:ListMountType( L["flying passenger"], info.handler.Flying.Passenger.Sets );
                       end,
            },
            [L["refresh"]] = {
                type = "execute",
                order = 15,
                name = L["Refresh mount list"],
                desc = L["Force a refresh of the list of all your mounts."],
                guiHidden = true,
                func = function( info )
                           info.handler.dirtyBags[0] = true;
                           info.handler.dirtyBags[1] = true;
                           info.handler.dirtyBags[2] = true;
                           info.handler.dirtyBags[3] = true;
                           info.handler.dirtyBags[4] = true;
                           info.handler.dirtyBags[-254] = true;
                           info.handler.dirtyBags[-255] = true;
                           info.handler:GetPlayersMounts();
                           info.handler:ChooseRandomMounts();
                           info.handler:SetButton( info.handler:GetStateBasedOnModifiers() );
                       end,
            },
            debug = {
                type = "toggle",
                hidden = true,
                name = "Debug",
                desc = "Debug mode toggle",
                desc = function( info )
                            local returnString = "Debug mode toggle";

                            if( info.uiType == "cmd" ) then
                                returnString = ("%s |cFFFF0000[%s]|r"):format( returnString, info.handler.DEBUG and L["yes"] or L["no"] );
                            end--if
                           return returnString;
                       end,
                get = function( info )
                          return info.handler.DEBUG;
                      end,
                set = function( info, val )
                          info.handler.DEBUG = val;
                          info.handler:Print( ("debug mode |cFFFF0000[%s]|r"):format( info.handler.DEBUG and L["yes"] or L["no"] ) );
                      end,
            },
        },
    };
    local function ProcessSets( addTo, sets, groupName, indexAdd )
        for _, name in pairs( sets ) do
            local speed = self.GetSetSpeed( name );
            
            if( speed ) then
                local index = speed + indexAdd;
                local sType = L["%1$i%% speed %2$s"]:format( speed, groupName );

                addTo[LibLordFarlander.NamifyString( name )] = {
                    type = "group",
                    handler = self,
                    order = index,
                    name = sType,
                    desc = sType,
                    args = typeTemplate,
                };
            end--if
        end--for
    end--ProcessSets( sets )
    local groundSets = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
    local groundPassengerSets = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Ground" ) or {};
    local flyingSets = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );
    local flyingPassengerSets = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Flying" ) or {};
    local BZ = LibStub( "LibBabble-Zone-3.0", true ):GetLookupTable();

    for continentindex, continent in pairs( { GetMapContinents() } ) do
        zoneOptions.args[LibLordFarlander.NamifyString( continent )] = {
            type = "group",
            handler = self,
            name = continent,
            desc = L["Zones in %s"]:format( continent ),
            childGroups = "tree",
            args = {},
        };

        local continentArgs = zoneOptions.args[LibLordFarlander.NamifyString( continent )].args;

        LibLordFarlander:CopyTable( continentTemplate, continentArgs );
        for index, zone in pairs( LibLordFarlander.GetMapZonesWithExtra( continentindex, true ) ) do
            continentArgs[LibLordFarlander.NamifyString( zone )] = {
                type = "group",
                handler = self,
                name = zone,
                desc = zone,
                childGroups = "tree",
                args = {},
            };
            
            local zoneArgs = continentArgs[LibLordFarlander.NamifyString( zone )].args;

            LibLordFarlander:CopyTable( zoneTemplate, zoneArgs );
            ProcessSets( zoneArgs, groundSets, L["ground"], 0 );
            ProcessSets( zoneArgs, groundPassengerSets, L["ground passenger"], 1000 );
            if( LibLordFarlander.IsFlyableZone( zone, true ) or ( zone == BZ["Dalaran"] ) ) then
                ProcessSets( zoneArgs, flyingSets, L["flying"], 10000 );
                ProcessSets( zoneArgs, flyingPassengerSets, L["flying passenger"], 11000 );
            end--if
        end--for
    end--for

    LibStub( "AceConfig-3.0" ):RegisterOptionsTable( L["Coconuts"], slashOptions, { L["coconuts"], L["ccn"] } );
    AceConfigReg:RegisterOptionsTable( "Coconuts Main Options", options );
    AceConfigReg:RegisterOptionsTable( "Coconuts Button", buttonOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Zone Favorites", zoneOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Favorites", favoritesOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Profile", profileOptions );
    AceConfigReg:RegisterOptionsTable( "Coconuts Modifer Items", modifierItemsOptions );
    AceConfigDialog:AddToBlizOptions( "Coconuts Main Options", L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Button", L["Button"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Modifer Items", L["Modifier Items"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Zone Favorites", L["Zone Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Favorites", L["Favorites"], L["Coconuts"] );
    AceConfigDialog:AddToBlizOptions( "Coconuts Profile", L["Profile"], L["Coconuts"] );
    
    if( LibStub( "LibDataBroker-1.1", true ) ) then
        self.ldbobj = LibStub( "LibDataBroker-1.1" ):NewDataObject( L["Coconuts"], {
                type = "data source",
                icon = "Interface\\AddOns\\Coconuts\\Img\\Coconuts",
                label = L["Coconuts"],
                text = L["You have no mounts."],
                secureTemplate = "SecureActionButtonTemplate",
                OnClick = function( button, btn )
                              if( button:IsObjectType( "SecureActionButtonTemplate" ) ) then
                                  self.button:Click( btn );
                              else
                                  if( self.mode == "Enchant" ) then
                                      self:FindNextMountEnchant();
                                  elseif( self.mode == "Dismount" ) then
                                      if( ( not IsFlying() ) or ( self.db.char.ModiferDismountFlying and (
                                              IsAltKeyDown() or IsControlKeyDown() or IsShiftKeyDown() ) or ( not self.db.char.ModiferDismountFlying ) ) ) then
                                          Dismount();
                                      end--if
                                  elseif( self.mode == "VehicleExit" ) then
                                      VehicleExit();
                                  else
                                      if( self.CurrentMountIndex ) then
                                          CallCompanion( "MOUNT", self.CurrentMountIndex );
                                      end--if
                                      self:ChooseRandomMounts();
                                  end--if
                                  self:SetButton( self:GetStateBasedOnModifiers() );
                              end--if
                          end,
                OnTooltipShow = function( tooltip )
                                    if( not tooltip ) then
                                        return;
                                    end--if
                                    self:SetTooltip( tooltip );
                                end,
            } );
    end--if
end--Coconuts:OnInitialize()

function Coconuts:OnProfileReset()
    self:OnProfileChanged();
end--Coconuts:OnProfileReset()

function Coconuts:OnProfileChanged()
    LibStub( "LibLordFarlander-UI-2.0" ).OnProfileChanged( self );
    profile = self.db.profile;
    if( profile.enchantOrder ) then
        local iterator, handle = PetAndMountDatabase:IterateSet( "Items.MountEnchant" );
        local item, value = iterator( handle );

        while item do
            local inList = false;

            for _, listItem in pairs( profile.enchantOrder ) do
                if( listItem == item ) then
                    inList = true;
                    break;
                end--if
            end--for
            if( not inList ) then
                tinsert( profile.enchantOrder, item );
            end--if
            item, value = iterator( handle );
        end--for
    else
        local iterator, handle = PetAndMountDatabase:IterateSet( "Items.MountEnchant" );
        local item, value = iterator( handle );
        local found = false;
        local counter = 1;

        TZ.ClearTable( profile.enchantOrder );
        while item do
            if( type( value ) == "boolean" ) then
                value = counter;
            else
                value = tonumber( value );
            end--if
            profile.enchantOrder[value] = item;
            item, value = iterator( handle );
            counter = counter + 1;
        end--for
    end--if
    --Upgrade check
    if( type( profile.LastVersion ) == "number" ) then
        --Upgrading from pre-2.5, need to convert zone favorites to the new format
        local newList = {};
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local GroundSets = PetAndMountDatabase:GetSubsets( "Mounts.Ground" );
        local FlyingSets = PetAndMountDatabase:GetSubsets( "Mounts.Flying" );

        for zone, itemID in pairs( profile.ZoneFavorites ) do
            newList[zone] = {};
            for _, set in pairs( GroundSets ) do
                if( ItemInSet( PetAndMountDatabase, itemID, set ) ) then
                    newList[zone][set] = { itemID = { exclusive = true, } };
                end--if
            end--if
            for _, set in pairs( FlyingSets ) do
                if( ItemInSet( PetAndMountDatabase, itemID, set ) ) then
                    newList[zone][set] = { itemID = { exclusive = true, } };
                end--if
            end--if
        end--for
        profile.ZoneFavorites = newList;
    end--if
    if( ( type( profile.LastWoWVersion ) ~= "number" ) or ( profile.LastWoWVersion < 30000 ) ) then
        --Upgrading from pre-wotlk, need to convert itemIDs in favorites and zone favorites to spellIDs
        local newList = {};
        local ItemInSet = PetAndMountDatabase.ItemInSet;

        for itemID, value in pairs( profile.Favorites ) do
            local newID = ItemInSet( PetAndMountDatabase, itemID, "Conversion.Mounts" );

            if( newID ) then
                itemID = tonumber( newID );
            end--if
            newList[itemID] = ( type( value ) == "number" ) and value or 2.0;
            if( newList[itemID] == 1.0 ) then
                newList[itemID] = nil;
            end--if
        end--for
        profile.Favorites = newList;
        newList = {};
        for zone, sets in pairs( profile.ZoneFavorites ) do
            newList[zone] = {};
            for set, mounts in pairs( sets ) do
                newList[zone][set] = {};
                for itemID, info in pairs( mounts ) do
                    local newID = ItemInSet( PetAndMountDatabase, itemID, "Conversion.Mounts" );

                    if( newID ) then
                        itemID = tonumber( newID );
                    end--if
                    newList[zone][set][itemID] = info;
                    if( not ( info.frequency or info.active or info.exclusive ) ) then
                        newList[zone][set][itemID] = nil;
                    end--if
                end--for
            end--for
        end--for
        profile.ZoneFavorites = newList;
    end--if
    profile.LastWoWVersion = LibLordFarlander.GetTOCVersion();
    profile.LastVersion = self.version;

    self.Dirty = true
    self:GetPlayersMounts();
    self:ChooseRandomMounts();
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:OnProfileChanged()

function Coconuts:OnEnable()
    --Register the Events
    local SpecialEvents_Mount = LibStub( "LibLordFarlander-SpecialEvents-Mount-4.0" );
    local SpecialEvents_Skill = LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0" );
    local LibUI = LibStub( "LibLordFarlander-UI-2.0" );

    self:RegisterEvent( "BAG_UPDATE",
        function( _, bagID )
            if( bagID and ( bagID >= 0 ) and ( bagID <= NUM_BAG_SLOTS ) ) then
                self.dirtyBags[bagID] = true;
                self.Dirty = true;
                if( not self.button:IsVisible() ) then
                    self:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "COMPANION_UPDATE",
        function( _, type )
            if( not type ) then
                self.Dirty = true;
                self.dirtyBags[-255] = true;
                if( not self.button:IsVisible() ) then
                    self:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    self:RegisterEvent( "MODIFIER_STATE_CHANGED",
        function()
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );
    self:RegisterEvent( "PLAYER_ALIVE",
        function()
            if( ( not self.Loaded ) or ( not self.RidingSkill ) ) then
                self.Loaded = true;
                self.RidingSkill = SpecialEvents_Skill:GetSkillLevel( L["Riding"] );
                self.dirtyBags = { [0] = true, [1] = true, [2] = true, [3] = true, [4] = true, [-254] = true, [-255] = true, };
                self:GetPlayersMounts();
                self:ChooseRandomMounts();
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_DISABLED",
        function()
            if( self.db.profile.Button.AutoHideButton and ( not ( self.db.profile.DefaultToFormInCombat and self.CurrentGroundForm ) ) ) then
                self.button:HideEx();
            elseif( self.db.profile.DefaultToFormInCombat and self.CurrentGroundForm ) then
                self.button:SetFromListEntry( self.CurrentGroundForm );
            end--if
        end );
    self:RegisterEvent( "PLAYER_REGEN_ENABLED",
        function()
            if( ( not self.Loaded ) or ( not self.RidingSkill ) ) then
                self.Loaded = true;
                self.RidingSkill = SpecialEvents_Skill:GetSkillLevel( L["Riding"] );
                self.dirtyBags = { [0] = true, [1] = true, [2] = true, [3] = true, [4] = true, [-254] = true, [-255] = true, };
                self:GetPlayersMounts();
                self:ChooseRandomMounts();
                self:SetButton( Coconuts:GetStateBasedOnModifiers() );
            end--if
            if( self.db.profile.Button.AutoHideButton ) then
                self.button:ShowEx();
            end--if
            self.button:OutOfCombatReset();
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );
    self:RegisterEvent( "UNIT_ENTERED_VEHICLE",
        function( _, unit )
            if( unit == "player" ) then
                self.InVehicle = true;
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    self:RegisterEvent( "UNIT_EXITED_VEHICLE",
        function( _, unit )
            if( unit == "player" ) then
                self.InVehicle = false;
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    self:RegisterEvent( "ZONE_CHANGED_NEW_AREA",
        function()
            if( self.db.profile.ZoneFavorites[GetRealZoneText()] ) then
                self:ChooseRandomMounts( true );
            end--if
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );

    SpecialEvents_Mount.RegisterCallback( self, "Mounted",
        function()
            if( not self.Mounted ) then
                self.Mounted = true;
                self:ChooseRandomMounts();
            end--if
        end );
    SpecialEvents_Mount.RegisterCallback( self, "Dismounted",
        function()
            if( self.Mounted ) then
                self.Mounted = false;
                self:SetButton( self:GetStateBasedOnModifiers() );
            end--if
        end );
    LibStub( "LibLordFarlander-SpecialEvents-LearnSpell-4.0" ).RegisterCallback( self, "LearnedSpell",
        function( _, spell )
            local spellID = LibLordFarlander.GetItemIDFromLink( GetSpellLink( spell ) );

            --Also check for learning Cold Weather Flying
            if( ( spellID ~= nil ) and ( ( spellID == LibLordFarlander.ColdWeatherFlying ) or PetAndMountDatabase:SpellInSet( spellID, "Mounts" ) or PetAndMountDatabase:SpellInSet( spellID, "Spells.Travel" ) ) ) then
                self.Dirty = true;
                if( spellID == LibLordFarlander.ColdWeatherFlying ) then
                    self.dirtyBags = { [0] = true, [1] = true, [2] = true, [3] = true, [4] = true, [-254] = true, [-255] = true, };
                else
                    self.dirtyBags[-255] = true;
                end--if
                if( not self.button:IsVisible() ) then
                    self:ScheduleTimer( "OnUpdate", 0 );
                end--if
            end--if
        end );
    SpecialEvents_Skill.RegisterCallback( self, "LearnedSkill_" .. L["Riding"],
        function( _, level )
            self.RidingSkill = level;
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );
    SpecialEvents_Skill.RegisterCallback( self, "SkillLeveled_" .. L["Riding"],
        function( _, level )
            self.RidingSkill = level;
            self:SetButton( self:GetStateBasedOnModifiers() );
        end );

    if( not self.button ) then
        LibUI.SetupForUI( self );
        self.button = LibUI.CreateButton( self, self.name, "Interface\\AddOns\\Coconuts\\Img\\Coconuts", self.db.profile.ButtonGroup, self.SkinChanged, self );
        ItemList.AttachToButton( self.button );
        self.button:SetMovable( not self.db.profile.Button.LockButton );
        self.button:HookScript( "OnClick",
            function( self, mouse )
                if( ( mouse == "RightButton" ) and ( self.class.mode == "Enchant" ) ) then
                    self.class:FindNextMountEnchant();
                end--if
            end );
    end--if

    --Check if upgrading from Mount
    if( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ) ) then
        SetBindingClick( GetBindingKey( "BINDING_NAME_CLICK Mount_Button:LeftButton" ), self.button:GetName(), "LeftButton" );
        SetBindingClick( nil, "Mount_Button", "LeftButton" );
    end--if

    self:OnProfileChanged();
end--Coconuts:OnEnable();

function Coconuts:OnDisable()
    LibStub( "LibLordFarlander-SpecialEvents-Mount-4.0" ).UnregisterAll( self );
    LibStub( "LibLordFarlander-SpecialEvents-LearnSpell-4.0" ).UnregisterAll( self );
    LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0" ).UnregisterAll( self );
end--Coconuts:OnDisable()

function Coconuts:AutoHideChange()
    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:AutoHideChange()

function Coconuts:OnUpdate( elapsed )
    local SwimCheck = self.CurrentAquaticForm and self.db.profile.SuggestForms;
    local oldSwimming = self.Swimming;

    self.Swimming = IsSwimming();
    if( self.Dirty ) then
        --Update the list of mount items if needed
        self.Dirty = false;        
        if( self:HasMountListChanged() ) then
            self:ChooseRandomMounts();
        end--if
        SwimCheck = false;
        self:SetButton( self:GetStateBasedOnModifiers() );
    end--if
    if( SwimCheck and ( oldSwimming ~= self.Swimming ) ) then
        self:SetButton( self:GetStateBasedOnModifiers() );
    end--if
end--Coconuts:OnUpdate()

function Coconuts:GetStateBasedOnModifiers()
    local profile = self.db.profile;
    local IsModifierDown = LibStub( "LibLordFarlander-UI-2.0" ).IsModifierDown;
    local IsFlyingModifierDown = IsModifierDown( profile.FlyingModifier );
    local IsPassengerModifierDown = IsModifierDown( profile.PassengerModifier );
    local IsMountItemModifierDown = IsModifierDown( profile.MountItemModifier );
    local IsNextBestModifierDown = IsModifierDown( profile.NextBestModifier );
    local BZ = LibStub( "LibBabble-Zone-3.0", true ):GetLookupTable();

    if( self.InVehicle ) then
        self.Mode = "VehicleExit";
        return nil;
    elseif( self.Mounted ) then
        if( profile.ShowModifierItems and IsMountItemModifierDown and self.CurrentMountItem ) then
            self.Mode = "Enchant";
            return self.CurrentMountItem;
        else
            self.Mode = "Dismount";
            return nil;
        end--if
    elseif( self.Swimming and self.CurrentAquaticForm and profile.SuggestForms ) then
        self.Mode = "Mount";
        return self.CurrentAquaticForm;
    else
        --Pick a mount or travel form
        if( LibLordFarlander.IsInFlyableZone() or ( ( GetRealZoneText() == BZ["Dalaran"] ) and ( GetSubZoneText() == L["Krasus' Landing"] ) and LibLordFarlander.HasPlayerSpell( LibLordFarlander.ColdWeatherFlying ) ) ) then
            if( ( profile.FavorFlying and ( not IsFlyingModifierDown ) ) or ( ( not profile.FavorFlying ) and IsFlyingModifierDown ) ) then
                if( ( profile.FavorPassenger and ( not IsPassengerModifierDown ) ) or ( ( not profile.FavorPassenger ) and IsPassengerModifierDown ) ) then
                    if( IsNextBestModifierDown ) then
                        if( self.NextCurrentFlyingPassenger ) then
                            self.Mode = "Mount";
                            return self.NextCurrentFlyingPassenger;
                        end--if
                    else
                        if( self.CurrentFlyingPassenger ) then
                            self.Mode = "Mount";
                            return self.CurrentFlyingPassenger;
                        end--if
                    end--if
                end--if
                if( IsNextBestModifierDown ) then
                    if( ( profile.SuggestForms and ( ( ( not profile.FlightFormDefault ) and self.CurrentFlying ) or self.CurrentFlyingForm ) ) or self.NextCurrentFlying ) then
                        self.Mode = "Mount";
                        return ( profile.SuggestForms and ( ( ( not profile.FlightFormDefault ) and self.CurrentFlying ) or self.CurrentFlyingForm ) ) or self.NextCurrentFlying;
                    end--if
                else
                    if( ( profile.SuggestForms and profile.FlightFormDefault and self.CurrentFlyingForm ) or self.CurrentFlying ) then
                        self.Mode = "Mount";
                        return ( profile.SuggestForms and profile.FlightFormDefault and self.CurrentFlyingForm ) or self.CurrentFlying;
                    end--if
                end--if
            end--if
        end--if
        if( ( profile.FavorPassenger and ( not IsPassengerModifierDown ) ) or ( ( not profile.FavorPassenger ) and IsPassengerModifierDown ) ) then
            if( IsNextBestModifierDown ) then
                if( self.NextCurrentGroundPassenger ) then
                    self.Mode = "Mount";
                    return ( profile.SuggestForms and self.CurrentGroundForm ) or self.NextCurrentGroundPassenger;
                end--if
            else
                if( self.CurrentGroundPassenger ) then
                    self.Mode = "Mount";
                    return self.CurrentGroundPassenger;
                end--if
            end--if
        end--if
        if( IsNextBestModifierDown ) then
            if( self.NextCurrentGround ) then
                self.Mode = "Mount";
                return ( profile.SuggestForms and self.CurrentGroundForm ) or self.NextCurrentGround;
            end--if
        else
            if( self.CurrentGround ) then
                self.Mode = "Mount";
                return self.CurrentGround;
            end--if
        end--if
    end--if
    self.Mode = "";
    return nil;
end--Coconuts:GetStateBasedOnModifiers()

----------------------------
--+Mount/Inventory Functions--
function Coconuts.GetSetSpeed( pt3set )
    local iterator, handle = PetAndMountDatabase:IterateSet( pt3set );
    local item, speed = iterator( handle );

    if( item and speed ) then
        speed = PetAndMountDatabase:GetMountAttributes( speed );
        return speed;
    end--if
    return nil
end--Coconuts.GetSetSpeed( pt3set )

function Coconuts:GetZoneMount( set, ZoneList, noFallBack )
    if( ZoneList and LibLordFarlander.HasTableMembers( ZoneList ) ) then
        local profile = self.db.profile;
        local favorites = TZ:GetTable();
        local exclusives = TZ:GetTable();
        local returnVal = nil;

        for mount, settings in pairs( ZoneList ) do
            if( settings.active ) then
                favorites[mount] = settings.frequency or 1.0;
            end--if
            if( settings.exclusive ) then
                exclusives[mount] = settings.frequency or 1.0;
            end--if
        end--if
        if( LibLordFarlander.HasTableMembers( favorites ) or LibLordFarlander.HasTableMembers( exclusives ) ) then
            local default = 1.0;

            if( LibLordFarlander.HasTableMembers( exclusives ) ) then
                TZ:ReturnTable( favorites );
                favorites = exclusives;
                exclusives = nil;
                default = 0.0;
            else
                local profileFavorites = self.db.profile.Favorites;

                for _, item in pairs( set ) do
                    if( not favorites[item] ) then
                        favorites[item] = profileFavorites[item] or 1.0;
                    end--if
                end--if
            end--if
            returnVal = self:GetRandomFromListWithWeighing( set, favorites, default );
        end--if
        TZ:ReturnTable( favorites );
        if( exclusives ) then
            TZ:ReturnTable( exclusives );
        end--if
        if( returnVal ) then
            return returnVal;
        end--if
    end--if
    return ( not noFallBack ) and self:GetRandomFromListWithWeighing( set, self.db.profile.Favorites ) or nil;
end--Coconuts:GetZoneMount( set, ZoneList, noFallBack )

function Coconuts:GetRandomFromListWithWeighing( set, favorites, default )
    if( not default ) then
        default = 1.0;
    end--if
    if( LibLordFarlander.HasTableMembers( favorites ) and set and ( LibLordFarlander.TableCount( set ) > 1 ) ) then
        local PlayerItemCount = LibLordFarlander.PlayerItemCount;
        local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;
        local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
        local HasPlayerSpell = LibLordFarlander.HasPlayerSpell;
        local ItemInSet = PetAndMountDatabase.ItemInSet;
        local profile = self.db.profile;
        local FindInBags = ItemList.FindInBags;
        local opSuggestOnlySoulbound = profile.SuggestOnlySoulbound;
        local opFavorMorePassengers = profile.FavorMorePassengers;
        local total = 0.0;
        local useset = TZ:GetTable();
        local randomValue;
        local amountOfItems = 0;
        local index = 0;
        local item = nil;

        for _, item in pairs( set ) do
            local reagented = nil;
            local itemID = item:GetID();
            local howOften = ( favorites[itemID] or default ) * ( opFavorMorePassengers and 1.0 or ( ( item.Passengers or 0.0 ) + 1.0 ) );
            local BagItem = ( itemID > 0 ) and FindInBags( itemID ) or nil;

            if( ( howOften > 0.0 ) and ( ( itemID < 0 ) or ( PlayerItemCount( itemID ) > 0 ) ) and ( ( itemID < 0 ) or ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( BagItem.bag, BagItem.slot ) ) ) ) then
                tinsert( useset, item );
                total = total + howOften;
                amountOfItems = amountOfItems + 1;
            end--if
        end--for
        randomValue = math.random() * total;
        repeat
            index = index + 1;
            item = useset[index];
            if( item ) then
                randomValue = randomValue - ( favorites[item:GetID()] or default ) * ( opFavorMorePassengers and 1.0 or ( ( item.Passengers or 0.0 ) + 1.0 ) );
            end--if
        until ( not item ) or ( randomValue <= 0 ) or ( index == amountOfItems );
        TZ:ReturnTable( useset );
        return item;
    else
        return ItemList.GetRandomItemFromList( set, current );
    end--if
end--Coconuts:GetRandomFromListWithWeighing( set, favorites, default )

function Coconuts:ChooseRandomMounts( noFallBack )
    local ZoneList = self.db.profile.ZoneFavorites[GetRealZoneText()] or {};

    self.CurrentGround = self:GetZoneMount( self:GetPetsList( self.Ground.Best ), ZoneList[self.Ground.Best], noFallBack ) or self.CurrentGround;
    self.NextCurrentGround = self:GetZoneMount( self:GetPetsList( self.Ground.NextBest ), ZoneList[self.Ground.NextBest], noFallBack ) or self.NextCurrentGround;
    self.CurrentFlying = self:GetZoneMount( self:GetPetsList( self.Flying.Best ), ZoneList[self.Flying.Best], noFallBack ) or self.CurrentFlying;
    self.NextCurrentFlying = self:GetZoneMount( self:GetPetsList( self.Flying.NextBest ), ZoneList[self.Flying.NextBest], noFallBack ) or self.NextCurrentFlying;
    
    self.CurrentGroundPassenger = self:GetZoneMount( self:GetPetsList( self.Ground.Passenger.Best ), ZoneList[self.Ground.Passenger.Best], noFallBack ) or self.CurrentGroundPassenger;
    self.NextCurrentGroundPassenger = self:GetZoneMount( self:GetPetsList( self.Ground.Passenger.NextBest ), ZoneList[self.Ground.Passenger.NextBest], noFallBack ) or self.NextCurrentGroundPassenger;
    self.CurrentFlyingPassenger = self:GetZoneMount( self:GetPetsList( self.Flying.Passenger.Best ), ZoneList[self.Flying.Passenger.Best], noFallBack ) or self.CurrentFlyingPassenger;
    self.NextCurrentFlyingPassenger = self:GetZoneMount( self:GetPetsList( self.Flying.Passenger.NextBest ), ZoneList[self.Flying.Passenger.NextBest], noFallBack ) or self.NextCurrentFlyingPassenger;

    self:SetButton( self:GetStateBasedOnModifiers() );
end--Coconuts:ChooseRandomMounts()

function Coconuts.GetBestSets( sets )
    local Best = nil;
    local BestName = "";
    local NextBest = nil;
    local NextBestName = "";

    for name, set in pairs( sets ) do
        if( Best == nil ) then
            Best = set;
            BestName = name;
        else
            if( set.speed > Best.speed ) then
                if( ( NextBest == nil ) or ( Best.speed > NextBest.speed ) ) then
                    NextBest = Best;
                    NextBestName = BestName;
                end--if
                Best = set;
                BestName = name;
            elseif( ( NextBest == nil ) or ( set.speed > NextBest.speed ) ) then
                NextBest = set;
                NextBestName = name;
            end--if
        end--if
    end--for
    if( not NextBest ) then
        NextBestName = BestName;
    end--if
    return BestName, NextBestName;
end--Coconuts.GetBestSets()

function Coconuts.AddIfMount( ptsets, dest, bag, slot, itemID )
    for _, set in pairs( ptsets ) do
        local speed = PetAndMountDatabase:ItemInSet( itemID, set );

        if( speed ) then
            local numspeed, riding, passengers = PetAndMountDatabase:GetMountAttributes( speed );
            local item;

            Coconuts:dprint( ("%i;%s;%s"):format( itemID, set, speed ) );
            if( dest[set] == nil ) then
                dest[set] = { bags = {}, ["speed"] = numspeed };
            end--if
            if( not dest[set].bags[bag] ) then
                dest[set].bags[bag] = {};
            end--if
            if( itemID < 0 ) then
                item = ItemList.AddSpellToList( dest[set].bags[bag], itemID * -1 );
            else
                item = ItemList.AddBagItemToList( dest[set].bags[bag], bag, slot );
            end--if
            if( item ) then
                item.Riding = riding;
                item.Passengers = passengers;
            end--if
            return true;
        end--if
    end--for
    return false;
end--Coconuts.AddIfMount( ptsets, dest, bag, slot, itemID )

function Coconuts.GetPlayerSpellMounts( spellDatabase, mountTable )
    local iterator, handle = PetAndMountDatabase:IterateSet( spellDatabase );

    if( iterator and handle ) then
        local spellID, speed = iterator( handle );

        while spellID do
            if( spellID < 0 ) then
                spellID = spellID * -1;
                if( LibLordFarlander.HasPlayerSpell( spellID ) ) then
                    local item;
                    local numspeed, riding, passengers = PetAndMountDatabase:GetMountAttributes( speed );

                    if( mountTable.speed == nil ) then
                        mountTable.speed = numspeed;
                    end--if
                    if( not mountTable.bags[-255] ) then
                        mountTable.bags[-255] = {};
                    end--if
                    item = ItemList.AddSpellToList( mountTable.bags[-255], spellID );
                    if( item ) then
                        item.Riding = riding;
                        item.Passengers = passengers;
                    end--if                    
                end--if
            end--if
            spellID, speed = iterator( handle );
        end--for
    end--if
end--Coconuts.GetPlayerSpellMounts( spellDatabase, mountTable )

function Coconuts.AddSpells( ptsets, dest, extra )
    for _, set in pairs( ptsets ) do
        local pt3set = set;

        if( extra ) then
            pt3set = ("%s.%s"):format( set, extra );
        end--if
        if( PetAndMountDatabase:HasSet( pt3set ) ) then
            local destSet = dest[set];
    
            if( destSet == nil ) then
                destSet = { bags = {}, };
            end--if
            Coconuts.GetPlayerSpellMounts( pt3set, destSet );
            if( destSet.speed ) then
                dest[set] = destSet;
            end--if
        end--if
    end--for
end--Coconuts.AddSpells( ptsets, dest, extra )

function Coconuts.AddIfVariableMount( GroundSets, FlyingSets, GroundPassengerSets, FlyingPassengerSets, bag, slot, itemID, RidingSkill )
    local ItemInSet = PetAndMountDatabase.ItemInSet;
    local isInAll = ItemInSet( PetAndMountDatabase, itemID, "Mounts.Variable.All" );
    local added = false;

    if( ( isInAll and ( RidingSkill <= 300 ) ) or ItemInSet( PetAndMountDatabase, itemID, "Mounts.Variable.Ground" ) ) then
        local destSet = "PetAndMountDatabase.Mounts.Ground.Fast";
        local speed = 60;

        if( RidingSkill >= 150 ) then
            speed = 100;
            destSet = "PetAndMountDatabase.Mounts.Ground.VeryFast";
        end--if
        if( GroundSets[destSet] == nil ) then
            GroundSets[destSet] = { bags = {}, ["speed"] = speed };
        end--if
        if( not GroundSets[destSet].bags[bag] ) then
            GroundSets[destSet].bags[bag] = {};
        end--if
        if( itemID < 0 ) then
            ItemList.AddSpellToList( GroundSets[destSet].bags[bag], itemID * -1 );
        else
            ItemList.AddBagItemToList( GroundSets[destSet].bags[bag], bag, slot );
        end--if
        added = true;
    end--if
    if( isInAll or ItemInSet( PetAndMountDatabase, spellID, "Mounts.Variable.Flying" ) ) then
        local destSet = "PetAndMountDatabase.Mounts.Flying.Fast";
        local speed = 60;

        if( RidingSkill >= 300 ) then
            speed = 280;
            destSet = "PetAndMountDatabase.Mounts.Flying.VeryFast";
        end--if
        if( FlyingSets[destSet] == nil ) then
            FlyingSets[destSet] = { bags = {}, ["speed"] = speed };
        end--if
        if( not FlyingSets[destSet].bags[bag] ) then
            FlyingSets[destSet].bags[bag] = {};
        end--if
        if( itemID < 0 ) then
            ItemList.AddSpellToList( FlyingSets[destSet].bags[bag], itemID * -1 );
        else
            ItemList.AddBagItemToList( FlyingSets[destSet].bags[bag], bag, slot );
        end--if
        added = true;
    end--if
    return added;
end--Coconuts.AddIfVariableMount( GroundSets, FlyingSets, bag, slot, itemID, RidingSkill )

function Coconuts.GetTravelForm( set )
    local iterator, handle = PetAndMountDatabase:IterateSet( ("Spells.Travel.%s"):format( set ) );
    local spell, _ = iterator( handle );

    while spell do
        spell = spell * -1;
        if( LibLordFarlander.HasPlayerSpell( spell ) ) then
            return ItemList.BuildSpell( spell );
        end--if
        spell, _ = iterator( handle );
    end--for
    return nil;
end--Coconuts.GetTravelForm( set )

function Coconuts.CleanupBags( sets, bag )
    for _, set in pairs( sets ) do
        if( set.bags ) then
            TZ.ClearTable( set.bags[bag] );
        end--if
    end--for
end--Coconuts.CleanupBags( set )

--Gets all your pets from your inventory.
function Coconuts:GetPlayersMounts()
    -- First do ground mountItems, best first
    local opSuggestOnlySoulbound = self.db.profile.SuggestOnlySoulbound;
    local dirtyBags = self.dirtyBags;
    local GroundMounts = PetAndMountDatabase:GetSubsets( "Mounts.Ground" ) or {};
    local FlyingMounts = PetAndMountDatabase:GetSubsets( "Mounts.Flying" ) or {};
    local GroundPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Ground" ) or {};
    local FlyingPassengerMounts = PetAndMountDatabase:GetSubsets( "Mounts.Passenger.Flying" ) or {};
    local GroundSets = self.Ground.Sets;
    local FlyingSets = self.Flying.Sets;
    local GroundPassengerSets = self.Ground.Passenger.Sets;
    local FlyingPassengerSets = self.Flying.Passenger.Sets;
    local ItemIsSoulbound = LibLordFarlander.ItemIsSoulbound;
    local PlayerMeetsMountRequirements = self.PlayerMeetsMountRequirements;
    local AddIfMount = self.AddIfMount;
    local AddIfVariableMount = self.AddIfVariableMount;
    local GetItemIDFromLink = LibLordFarlander.GetItemIDFromLink;
    local CleanupBags = self.CleanupBags;

    self.CurrentMountItem = nil;
    self.Ground.Best = "";
    self.Ground.NextBest = "";
    self.Ground.Passenger.Best = "";
    self.Ground.Passenger.NextBest = "";
    self.Flying.Best = "";
    self.Flying.NextBest = "";
    self.Flying.Passenger.Best = "";
    self.Flying.Passenger.NextBest = "";
    
    if( not self.RidingSkill ) then
        self.RidingSkill = LibStub( "LibLordFarlander-SpecialEvents-Skill-4.0" ):GetSkillLevel( L["Riding"] );
    end--if

    for bag = 0, NUM_BAG_SLOTS do
        if( dirtyBags[bag] ) then
            CleanupBags( GroundSets, bag );
            CleanupBags( FlyingSets, bag );
            CleanupBags( GroundPassengerSets, bag );
            CleanupBags( FlyingPassengerSets, bag );
            for slot = 1, GetContainerNumSlots( bag ) do
                local itemLink = GetContainerItemLink( bag, slot );
    
                if( itemLink ) then
                    local itemID = GetItemIDFromLink( itemLink );

                    if( itemID and PlayerMeetsMountRequirements( self, itemID ) and ( ( not opSuggestOnlySoulbound ) or ItemIsSoulbound( bag, slot ) ) ) then
                        AddIfMount( GroundMounts, GroundSets, bag, slot, itemID );
                        AddIfMount( FlyingMounts, FlyingSets, bag, slot, itemID );
                        AddIfMount( GroundPassengerMounts, GroundPassengerSets, bag, slot, itemID );
                        AddIfMount( FlyingPassengerMounts, FlyingPassengerSets, bag, slot, itemID );
                        AddIfVariableMount( GroundSets, FlyingSets, GroundPassengerSets, FlyingPassengerSets, bag, slot, itemID, self.RidingSkill or 0 );
                    end--if
                end--if
            end--for
        end--if
    end--for

    if( dirtyBags[-255] ) then
        local gratuity = LibStub( "LibGratuity-3.0", true );

        CleanupBags( GroundSets, -255 );
        CleanupBags( FlyingSets, -255 );
        CleanupBags( GroundPassengerSets, -255 );
        CleanupBags( FlyingPassengerSets, -255 );

        for id, mount in pairs( PetAndMountDatabase:GetPlayersCompanions( "MOUNT" ) ) do
            local spellID = mount.spellID * -1;
            local added = false;                

            added = AddIfMount( GroundMounts, GroundSets, -255, nil, spellID ) or added;
            added = AddIfMount( FlyingMounts, FlyingSets, -255, nil, spellID ) or added;
            added = AddIfMount( GroundPassengerMounts, GroundPassengerSets, -255, nil, spellID ) or added;
            added = AddIfMount( FlyingPassengerMounts, FlyingPassengerSets, -255, nil, spellID ) or added;
            added = AddIfVariableMount( GroundSets, FlyingSets, GroundPassengerSets, FlyingPassengerSets, -255, nil, spellID, self.RidingSkill or 0 ) or added;
            if( not added ) then
                --we KNOW this a mount, but the database is out of date and doesn't have it
                --we'll have to use the tooltip to figure out if it's flying or not and its speed
                --and if it can take passengers
                self:dprint( ("Could not identify %s [spellID=%i;creatureID=%i]!"):format( mount.name, mount.spellID, mount.modelID ) );
                spellID = spellID * -1;
                gratuity:SetHyperlink( GetSpellLink( spellID ) );

                local strings = gratuity:GetText( 3, 4, false, true );

                if( strings ) then
                    local superType = "";
                    local type = "Ground";
                    local speedText = "Fast";
                    local speed = 60;
                    local destSet;
                    local sets;

                    for _, text in pairs( strings ) do
                        text = text[1];
                        if( text:find( L["[Cc]an only be [%w]+ed in [%w%s]-%."] ) ) then
                            type = "Flying";
                        end--if
                        if( text:find( L["This is a fast [%w%s]-[%.!]"] ) ) then
                            speed = 60;
                            speedText = "Fast";
                        end--if
                        if( text:find( L["This is a very fast [%w%s]-[%.!]"] ) ) then
                            if( type == "Flying" ) then
                                speed = 280;
                            else
                                speed = 100;
                            end--if
                            speedText = "VeryFast";
                        end--if
                        if( text:find( L["This is an extremely fast [%w%s]-[%.!]"] ) ) then
                            speed = 310;
                            speedText = "ExtremelyFast";
                        end--if
                        if( text:find( L["This [%w%s]- changes depending on your Riding skill%."] ) ) then
                            --Variable speed, flying or ground only
                            if( type == "Flying" ) then
                                if( ( self.RidingSkill or 0 ) >= 300 ) then
                                    speed = 280;
                                    speedText = "VeryFast";
                                else
                                    speed = 60;
                                    speedText = "Fast";                                    
                                end--if
                            else
                                if( ( self.RidingSkill or 0 ) >= 150 ) then
                                    speed = 100;
                                    speedText = "VeryFast";
                                else
                                    speed = 60;
                                    speedText = "Fast";                                    
                                end--if
                            end--if
                        end--if
                        if( text:find( L["can carry (%d+) passengers"] ) or text:find( L["complete with two buddies"] ) or text:find( L["Allows [%w%s]- passengers to board."] ) ) then
                            superType = "Passenger";
                        end--if
                        if( text:find( L["This [%w%s]- changes depending on your Riding skill and location%."] ) ) then
                            --whoo!  This mount is a ground mount in Azeroth, flying otherwise
                            --Also the speed depends on the player's riding skill!
                            if( ( self.RidingSkill or 0 ) >= 300 ) then
                                local additionalSet = ("PetAndMountDatabase.Mounts%sGround.VeryFast"):format( "." .. superType );

                                speed = 280;
                                speedText = "VeryFast";
                                type = "Flying";
                                --Also add to very fast ground mounts
                                sets = self.Ground.Sets;
                                if( sets[additionalSet] == nil ) then
                                    sets[additionalSet] = { bags = {}, ["speed"] = 100 };
                                end--if
                                if( not sets[additionalSet].bags[-255] ) then
                                    sets[additionalSet].bags[-255] = {};
                                end--if
                                ItemList.AddSpellToList( sets[additionalSet].bags[-255], spellID );
                            elseif( ( self.RidingSkill or 0 ) >= 225 ) then
                                local additionalSet = ("PetAndMountDatabase.Mounts%sGround.VeryFast"):format( "." .. superType );

                                speed = 60;
                                speedText = "Fast";
                                type = "Flying";
                                --Also add to very fast ground mounts
                                sets = self.Ground.Sets;
                                if( sets[additionalSet] == nil ) then
                                    sets[additionalSet] = { bags = {}, ["speed"] = 100 };
                                end--if
                                if( not sets[additionalSet].bags[-255] ) then
                                    sets[additionalSet].bags[-255] = {};
                                end--if
                                ItemList.AddSpellToList( sets[additionalSet].bags[-255], spellID );
                            elseif( ( self.RidingSkill or 0 ) >= 150 ) then
                                speed = 100;
                                speedText = "VeryFast";
                                type = "Ground";
                            elseif( ( self.RidingSkill or 0 ) >= 75 ) then
                                speed = 60;
                                speedText = "Fast";
                                type = "Ground";
                            end--if
                        end--if
                    end--for
                    destSet = ("PetAndMountDatabase.Mounts%s%s.%s"):format( "." .. superType, type, speedText );
                    if( superType == "" ) then
                        sets = self[type].Sets;
                    else
                        sets = self[type][superType].Sets;
                    end--if
                    if( sets[destSet] == nil ) then
                        sets[destSet] = { bags = {}, ["speed"] = speed };
                    end--if
                    if( not sets[destSet].bags[-255] ) then
                        sets[destSet].bags[-255] = {};
                    end--if
                    ItemList.AddSpellToList( sets[destSet].bags[-255], spellID );
                end--if
            end--if
        end--for
    end--if

    self.Ground.Best, self.Ground.NextBest = self.GetBestSets( GroundSets );
    self.Flying.Best, self.Flying.NextBest = self.GetBestSets( FlyingSets );
    self.Ground.Passenger.Best, self.Ground.Passenger.NextBest = self.GetBestSets( GroundPassengerSets );
    self.Flying.Passenger.Best, self.Flying.Passenger.NextBest = self.GetBestSets( FlyingPassengerSets );

    self.CurrentMountItemIndex = nil;
    self:FindNextMountEnchant();

    self.CurrentGroundForm = self.GetTravelForm( "Ground" );
    self.CurrentFlyingForm = self.GetTravelForm( "Flying" );
    self.CurrentAquaticForm = self.GetTravelForm( "Aquatic" );
    TZ.ClearTable( self.dirtyBags );
    if( not self.RidingSkill ) then
        self.dirtyBags[-255] = true;
    end--if
end--Coconuts:GetPlayersMounts()

function Coconuts:PrintMountList( mountTable, string )
    if( #mountTable > 0 ) then
        self:Print( L["%s mounts:"]:format( string ) );
        for _, mount in pairs( mountTable ) do
            self:Print( ("|T%s:0|t %s"):format( mount:GetTexture(), mount:GetLink() ) );
        end--for
    end--if
end--Coconuts:PrintMountList()

function Coconuts:ListMountType( type, lists )
    local sorted = TZ:GetTable();

    for name, _ in pairs( lists ) do
        tinsert( sorted, name );
    end--for
    table.sort( sorted, function( a, b ) return lists[a].speed < lists[b].speed end );
    for _, name in pairs( sorted ) do
        self:PrintMountList( self:GetPetsList( name ), L["%1$i%% speed %2$s"]:format( lists[name].speed, type ) );
    end--for
    TZ:ReturnTable( sorted );
end--Coconuts.ListMountType( type, list )

function Coconuts:GetPetsList( set )
    local returnVal = {};

    if( set:find( "%.Passenger" ) ) then
        if( set:find( "%.Ground" ) ) then
            if( self.Ground.Passenger.Sets[set] and self.Ground.Passenger.Sets[set].bags ) then
                for _, bag in pairs( self.Ground.Passenger.Sets[set].bags ) do
                    for _, mount in pairs( bag ) do
                        tinsert( returnVal, mount );
                    end--for
                end--for
            end--if
            return returnVal;
        end--if
        if( set:find( "%.Flying" ) ) then
            if( self.Flying.Passenger.Sets[set] and self.Flying.Passenger.Sets[set].bags ) then
                for _, bag in pairs( self.Flying.Passenger.Sets[set].bags ) do
                    for _, mount in pairs( bag ) do
                        tinsert( returnVal, mount );
                    end--for
                end--for
            end--if
            return returnVal;
        end--if
    else
        if( set:find( "%.Ground" ) ) then
            if( self.Ground.Sets[set] and self.Ground.Sets[set].bags ) then
                for _, bag in pairs( self.Ground.Sets[set].bags ) do
                    for _, mount in pairs( bag ) do
                        tinsert( returnVal, mount );
                    end--for
                end--for
            end--if
            return returnVal;
        end--if
        if( set:find( "%.Flying" ) ) then
            if( self.Flying.Sets[set] and self.Flying.Sets[set].bags ) then
                for _, bag in pairs( self.Flying.Sets[set].bags ) do
                    for _, mount in pairs( bag ) do
                        tinsert( returnVal, mount );
                    end--for
                end--for
            end--if
            return returnVal;
        end--if
    end--if
    return returnVal;
end--Coconuts.GetPetsList( set )

function Coconuts.GetSetMountsList( sets, returnList, used )
    if( not used ) then
        used = TZ:GetTable();
    end--if
    for _, set in pairs( sets ) do
        for _, bag in pairs( set.bags ) do
            for _, mount in pairs( bag ) do
                if( not used[mount:GetID()] ) then
                    tinsert( returnList, mount );
                    used[mount:GetID()] = true;
                end--if
            end--for
        end--for
    end--for
end--Coconuts.GetSetMountsList( set, returnList, used )

function Coconuts:GetAllMountsList()
    local returnList = {};
    local used = TZ:GetTable();
    
    self.GetSetMountsList( self.Ground.Sets, returnList, used );
    self.GetSetMountsList( self.Flying.Sets, returnList, used );
    self.GetSetMountsList( self.Ground.Passenger.Sets, returnList, used );
    self.GetSetMountsList( self.Flying.Passenger.Sets, returnList, used );
    TZ:ReturnTable( used );
    return returnList;
end--Coconuts:GetAllMountsList()
---Coconuts/Inventory Functions--
----------------------------

-------------------
--+Help Functions--
function Coconuts:PlayerMeetsMountRequirements( itemID )
    local speed = PetAndMountDatabase:ItemInSet( itemID, "Mounts" );

    if( speed ) then
        local itemMinLevel = select( 5, GetItemInfo( itemID ) ) or 0;
        local ridingSkill = tonumber( select( 2, strsplit( "|", speed ) ) or 0 );

        return ( itemMinLevel <= UnitLevel( "player" ) ) and ( ridingSkill <= ( self.RidingSkill or 0 ) );
    end--if
    return false;
end--if

function Coconuts:HasMountListChanged()
    local selfBestGround = self.Ground.Sets[self.Ground.Best];
    local selfBestFlying = self.Flying.Sets[self.Flying.Best];
    local selfBestPassengerGround = self.Ground.Passenger.Sets[self.Ground.Passenger.Best];
    local selfBestPassengerFlying = self.Flying.Passenger.Sets[self.Flying.Passenger.Best];
    local selfNextBestGround = self.Ground.Sets[self.Ground.NextBest];
    local selfNextBestFlying = self.Flying.Sets[self.Flying.NextBest];
    local selfNextBestPassengerGround = self.Ground.Passenger.Sets[self.Ground.Passenger.NextBest];
    local selfNextBestPassengerFlying = self.Flying.Passenger.Sets[self.Flying.Passenger.NextBest];
    local preGroundCount = #selfBestGround;
    local preFlyingCount = #selfBestFlying;
    local prePassengerGroundCount = #selfBestPassengerGround;
    local prePassengerFlyingCount = #selfBestPassengerFlying;
    local preNextGroundCount = #selfNextBestGround;
    local preNextFlyingCount = #selfNextBestFlying;
    local preNextPassengerGroundCount = #selfNextBestPassengerGround;
    local preNextPassengerFlyingCount = #selfNextBestPassengerFlying;
    local preBestGround = TZ:GetTable();
    local preBestFlying = TZ:GetTable();
    local preBestPassengerGround =  TZ:GetTable();
    local preBestPassengerFlying = TZ:GetTable();
    local preNextBestGround =  TZ:GetTable();
    local preNextBestFlying = TZ:GetTable();
    local preNextBestPassengerGround =  TZ:GetTable();
    local preNextBestPassengerFlying = TZ:GetTable();
    local returnVal = false;

    LibLordFarlander.CopyTable( preBestGround, selfBestGround );
    LibLordFarlander.CopyTable( preBestFlying, selfBestFlying );
    LibLordFarlander.CopyTable( preBestPassengerGround, selfBestPassengerGround );
    LibLordFarlander.CopyTable( preBestPassengerFlying, selfBestPassengerFlying );
    LibLordFarlander.CopyTable( preNextBestGround, selfNextBestGround );
    LibLordFarlander.CopyTable( preNextBestFlying, selfNextBestFlying );
    LibLordFarlander.CopyTable( preNextBestPassengerGround, selfNextBestPassengerGround );
    LibLordFarlander.CopyTable( preNextBestPassengerFlying, selfNextBestPassengerFlying );

    self:GetPlayersMounts();
    
    selfBestGround = self.Ground.Sets[self.Ground.Best];
    selfBestFlying = self.Flying.Sets[self.Flying.Best];
    selfBestPassengerGround = self.Ground.Passenger.Sets[self.Ground.Passenger.Best];
    selfBestPassengerFlying = self.Flying.Passenger.Sets[self.Flying.Passenger.Best];
    selfNextBestGround = self.Ground.Sets[self.Ground.NextBest];
    selfNextBestFlying = self.Flying.Sets[self.Flying.NextBest];
    selfNextBestPassengerGround = self.Ground.Passenger.Sets[self.Ground.Passenger.NextBest];
    selfNextBestPassengerFlying = self.Flying.Passenger.Sets[self.Flying.Passenger.NextBest];

    if( not self.ReadyToInitialize ) then
        self.ReadyToInitialize = true;
        return true;
    end
    self.ReadyToInitialize = true;
    returnVal = not( LibLordFarlander.CompareTables( preBestGround, selfBestGround ) and LibLordFarlander.CompareTables( preBestFlying, selfBestFlying ) and
        LibLordFarlander.CompareTables( preNextBestGround, selfNextBestGround ) and LibLordFarlander.CompareTables( preNextBestFlying, selfNextBestFlying ) and
        LibLordFarlander.CompareTables( preBestPassengerGround, selfBestPassengerGround ) and LibLordFarlander.CompareTables( preBestPassengerFlying, selfBestPassengerFlying ) and
        LibLordFarlander.CompareTables( preNextBestPassengerGround, selfNextBestPassengerGround ) and LibLordFarlander.CompareTables( preNextBestPassengerFlying, selfNextBestPassengerFlying ) );
    TZ:ReturnTable( preBestGround );
    TZ:ReturnTable( preBestFlying );
    TZ:ReturnTable( preBestPassengerGround );
    TZ:ReturnTable( preBestPassengerFlying );
    TZ:ReturnTable( preNextBestGround );
    TZ:ReturnTable( preNextBestFlying );
    TZ:ReturnTable( preNextBestPassengerGround );
    TZ:ReturnTable( preNextBestPassengerFlying );
    return returnVal;
end--Coconuts:HasMountListChanged()

function Coconuts:FindNextMountEnchant()
    local enchantOrder = self.db.profile.enchantOrder;
    local index = ( self.CurrentMountItemIndex or 0 ) + 1;
    local total = #enchantOrder;
    local count = 0;

    if( index > #enchantOrder ) then
        index = 1;
    end--if
    while( count < total ) do
        if( GetItemCount( enchantOrder[index] ) > 0 ) then
            self.CurrentMountItem = ItemList.BuildItem( enchantOrder[index] );
            self.CurrentMountItemIndex = index;
            return;
        end--if
        index = index + 1;
        if( index > #enchantOrder ) then
            index = 1;
        end--if
        count = count + 1;
    end--while
    self.CurrentMountItem = nil;
    self.CurrentMountItemIndex = nil;
end--Coconuts:FindNextMountEnchant()

function Coconuts:QuickSummon( mountID )
    local companionID = PetAndMountDatabase:FindCompanion( "MOUNT", mountID )
    if( companionID ) then
       CallCompanion( "MOUNT", companionID );
       self:ChooseRandomMounts();
       self:SetButton( self:GetStateBasedOnModifiers() );
    end--if
end--Coconuts:QuickSummon( mountID )
---Help Functions--
-------------------

-----------------
--+UI Functions--
function Coconuts:SetTooltip( tooltip )
    if( self.Mode == "Dismount" ) then
        local chardb = self.db.profile;

        if( chardb.ModiferDismountFlying ) then
            if( chardb.ShowModifierItems ) then
                local notUsed = TZ:GetTable();
                local used = self.db.profile.MountItemModifier;
                
                if( used ~= "Alt" ) then
                    tinsert( notUsed, L["Alt"] );
                end--if
                if( used ~= "Control" ) then
                    tinsert( notUsed, L["Control"] );
                end--if
                if( used ~= "Shift" ) then
                    tinsert( notUsed, L["Shift"] );
                end--if
                tooltip:SetText( L["Dismount (if flying hold down %s or %s and click to dismount)"]:format( notUsed[1], notUsed[2] ) );
                TZ:ReturnTable( notUsed );
            else
                tooltip:SetText( L["Dismount (if flying hold down Shift, Control, or Alt and click to dismount)"] );
            end--if
        else
            tooltip:SetText( L["Dismount"] );
        end--if
    elseif( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) then
        ItemList.SetTooltipFromListEntry( tooltip, self.CurrentMount );
    elseif( self.Mode == "VehicleExit" ) then
        tooltip:SetText( L["Exit Vehicle"] );
    else
        tooltip:SetText( L["You have no mounts."] );
    end--if
end--Coconuts:SetTooltip( tooltip )

function Coconuts:SetGameTooltips()
    local mntButton = self.button;
    local mntTooltip = mntButton.tooltip;

    if( ( not self.keyBoundMode ) and ( not InCombatLockdown() ) ) then
        mntTooltip:SetOwner( mntButton, "RIGHT" );
        self:SetTooltip( mntTooltip );
    elseif( InCombatLockdown() ) then
        mntTooltip:SetText( L["Cannot mount, you are in combat!"] );
    end--if
    LibStub( "LibLordFarlander-UI-2.0" ).AfterSetTooltip( mntButton );
end--Coconuts:SetGameTooltips()

function Coconuts:SetButton( mountItem )
    if( not InCombatLockdown() ) then
        local mntButton = self.button;
        local chardb = self.db.profile;
        local ldbobj = self.ldbobj;

        if( chardb.Button.AutoHideButton and mountItem ) then
            mntButton:ShowEx();
        end--if
        if( self.Mode == "Dismount" ) then
            mntButton.cooldown:Hide();
            mntButton:ShowEx();
            mntButton:Enable();
            mntButton:SetAttribute( "*type1", "macro" );
            mntButton:SetAttribute( "macrotext", chardb.ModiferDismountFlying and "/dismount [noflying][mod]" or "/dismount" );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\Dismount" );

            ldbobj.icon = "Interface\\AddOns\\Coconuts\\Img\\Dismount";
            ldbobj.text = L["Dismount"];
        elseif( self.Mode == "VehicleExit" ) then
            mntButton.cooldown:Hide();
            mntButton:ShowEx();
            mntButton:Enable();
            mntButton:SetAttribute( "*type1", "macro" );
            mntButton:SetAttribute( "macrotext", "/script VehicleExit();" );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\VehicleExit" );

            ldbobj.icon = "Interface\\AddOns\\Coconuts\\Img\\VehicleExit";
            ldbobj.text = L["Exit Vehicle"];
        elseif( mountItem and ( ( self.Mode == "Mount" ) or ( self.Mode == "Enchant" ) ) ) then
            mntButton:Enable();
            mntButton:SetFromListEntry( mountItem, not LibLordFarlander.CompareTables( self.CurrentMount, mountItem ) );
            self.CurrentMountIndex = PetAndMountDatabase:FindCompanion( "MOUNT", mountItem.spellID );
        else
            if( chardb.Button.AutoHideButton ) then
                mntButton:HideEx();
            end--if
            mntButton.cooldown:Hide();
            mntButton:SetAttribute( "*type1", "item" );
            mntButton:SetAttribute( "*bag1", nil );
            mntButton:SetAttribute( "*slot1", nil );
            mntButton:SetAttribute( "*item1", nil );
            mntButton:SetImage( "Interface\\AddOns\\Coconuts\\Img\\Coconuts" );

            ldbobj.icon = "Interface\\AddOns\\Coconuts\\Img\\Coconuts";
            ldbobj.text = L["You have no mounts."];
        end--if
        self.CurrentMount = mountItem;
    end--if
end--Coconuts:SetButton( mountItem )
---UI Functions--
-----------------
